@@ -36,14 +36,11 @@
 //!     use background_jobs::memory_storage::Storage;
 //!     let storage = Storage::new();
 //!
-//!     // Start the application server. This guards access to to the jobs store
-//!     let queue_handle = create_server(storage);
-//!
 //!     // Configure and start our workers
-//!     WorkerConfig::new(move || MyState::new("My App"))
+//!     let queue_handle = WorkerConfig::new(storage, move |_| MyState::new("My App"))
 //!         .register::<MyJob>()
 //!         .set_worker_count(DEFAULT_QUEUE, 16)
-//!         .start(queue_handle.clone());
+//!         .start();
 //!
 //!     // Queue our jobs
 //!     queue_handle.queue(MyJob::new(1, 2))?;
@@ -120,66 +117,269 @@ use actix_rt::{Arbiter, ArbiterHandle};
 use anyhow::Error;
 use background_jobs_core::{new_job, new_scheduled_job, Job, ProcessorMap, Stats, Storage};
 use chrono::{DateTime, Utc};
-use std::{collections::BTreeMap, sync::Arc, time::Duration};
-use tracing::error;
+use std::{collections::BTreeMap, marker::PhantomData, ops::Deref, sync::Arc, time::Duration};
 
 mod every;
 mod server;
 mod storage;
 mod worker;
 
-use self::{every::every, server::Server, worker::local_worker};
+use self::{every::every, server::Server, worker::LocalWorkerStarter};
 
 pub use background_jobs_core::ActixJob;
 
+/// Manager for worker threads
+///
+/// Manager attempts to restart workers as their arbiters die
+pub struct Manager {
+    // the manager arbiter
+    _arbiter: ArbiterDropper,
+
+    // handle for queueing
+    queue_handle: QueueHandle,
+}
+
+struct ArbiterDropper {
+    arbiter: Option<Arbiter>,
+}
+
+impl Manager {
+    /// Create a new manager to keep jobs alive
+    ///
+    /// Manager works by startinng a new Arbiter to run jobs, and if that arbiter ever dies, it
+    /// spins up another one and spawns the workers again
+    fn new<State>(worker_config: WorkerConfig<State, Managed>) -> Self
+    where
+        State: Clone,
+    {
+        let arbiter = Arbiter::new();
+        let worker_arbiter = Arbiter::new();
+        let notifier = Arc::new(tokio::sync::Notify::new());
+
+        let queue_handle = worker_config.queue_handle.clone();
+
+        let drop_notifier = DropNotifier::new(Arc::clone(&notifier));
+        arbiter.spawn(async move {
+            let mut drop_notifier = drop_notifier;
+            let mut arbiter = ArbiterDropper {
+                arbiter: Some(worker_arbiter),
+            };
+
+            loop {
+                worker_config
+                    .queue_handle
+                    .inner
+                    .ticker(arbiter.handle(), drop_notifier.clone());
+                worker_config.start_managed(&arbiter.handle(), &drop_notifier);
+
+                notifier.notified().await;
+                // drop_notifier needs to live at least as long as notifier.notified().await
+                // in order to ensure we get notified by ticker or a worker, and not ourselves
+                drop(drop_notifier);
+
+                // Assume arbiter is dead if we were notified
+                let online = arbiter.spawn(async {});
+                if online {
+                    panic!("Arbiter should be dead by now");
+                }
+
+                arbiter = ArbiterDropper {
+                    arbiter: Some(Arbiter::new()),
+                };
+
+                drop_notifier = DropNotifier::new(Arc::clone(&notifier));
+            }
+        });
+
+        Manager {
+            _arbiter: ArbiterDropper {
+                arbiter: Some(arbiter),
+            },
+            queue_handle,
+        }
+    }
+
+    /// Retrieve the QueueHandle for the managed workers
+    pub fn queue_handle(&self) -> &QueueHandle {
+        &self.queue_handle
+    }
+}
+
+impl Deref for Manager {
+    type Target = QueueHandle;
+
+    fn deref(&self) -> &Self::Target {
+        &self.queue_handle
+    }
+}
+
+impl Deref for ArbiterDropper {
+    type Target = Arbiter;
+
+    fn deref(&self) -> &Self::Target {
+        self.arbiter.as_ref().unwrap()
+    }
+}
+
+impl Drop for ArbiterDropper {
+    fn drop(&mut self) {
+        self.stop();
+        let _ = self.arbiter.take().unwrap().join();
+    }
+}
+
+#[derive(Clone)]
+struct DropNotifier {
+    inner: Arc<std::sync::Mutex<Option<Arc<tokio::sync::Notify>>>>,
+}
+
+impl DropNotifier {
+    fn new(notify: Arc<tokio::sync::Notify>) -> Self {
+        DropNotifier {
+            inner: Arc::new(std::sync::Mutex::new(Some(notify))),
+        }
+    }
+}
+
+impl Drop for DropNotifier {
+    fn drop(&mut self) {
+        if let Some(notifier) = self.inner.lock().unwrap().take() {
+            notifier.notify_one();
+        }
+    }
+}
+
 /// Create a new Server
 ///
 /// In previous versions of this library, the server itself was run on it's own dedicated threads
 /// and guarded access to jobs via messages. Since we now have futures-aware synchronization
 /// primitives, the Server has become an object that gets shared between client threads.
+fn create_server_in_arbiter<S>(arbiter: ArbiterHandle, storage: S) -> QueueHandle
+where
+    S: Storage + Sync + 'static,
+{
+    let handle = create_server_managed(storage);
+    handle.inner.ticker(arbiter, ());
+    handle
+}
+
+/// Create a new managed Server
 ///
-/// This method will panic if not called from an actix runtime
-pub fn create_server<S>(storage: S) -> QueueHandle
+/// In previous versions of this library, the server itself was run on it's own dedicated threads
+/// and guarded access to jobs via messages. Since we now have futures-aware synchronization
+/// primitives, the Server has become an object that gets shared between client threads.
+fn create_server_managed<S>(storage: S) -> QueueHandle
 where
     S: Storage + Sync + 'static,
 {
-    let arbiter = Arbiter::current();
     QueueHandle {
-        inner: Server::new(&arbiter, storage),
-        arbiter,
+        inner: Server::new(storage),
     }
 }
 
+/// Marker type for Unmanaged workers
+pub struct Unmanaged;
+/// Marker type for Managed workers
+pub struct Managed;
+
 /// Worker Configuration
 ///
 /// This type is used for configuring and creating workers to process jobs. Before starting the
 /// workers, register `Job` types with this struct. This worker registration allows for
 /// different worker processes to handle different sets of workers.
 #[derive(Clone)]
-pub struct WorkerConfig<State>
+pub struct WorkerConfig<State, M>
 where
     State: Clone + 'static,
 {
     processors: ProcessorMap<State>,
     queues: BTreeMap<String, u64>,
+    arbiter: Option<ArbiterHandle>,
+    queue_handle: QueueHandle,
+    managed: PhantomData<M>,
 }
 
-impl<State> WorkerConfig<State>
+impl<State> WorkerConfig<State, Managed>
 where
     State: Clone + 'static,
 {
-    /// Create a new WorkerConfig
+    /// Create a new managed WorkerConfig
     ///
     /// The supplied function should return the State required by the jobs intended to be
     /// processed. The function must be sharable between threads, but the state itself does not
     /// have this requirement.
-    pub fn new(state_fn: impl Fn() -> State + Send + Sync + 'static) -> Self {
+    pub fn new_managed<S: Storage + Send + Sync + 'static>(
+        storage: S,
+        state_fn: impl Fn(QueueHandle) -> State + Send + Sync + 'static,
+    ) -> Self {
+        let queue_handle = create_server_managed(storage);
+        let q2 = queue_handle.clone();
+
         WorkerConfig {
-            processors: ProcessorMap::new(Arc::new(state_fn)),
+            processors: ProcessorMap::new(Arc::new(move || state_fn(q2.clone()))),
             queues: BTreeMap::new(),
+            arbiter: None,
+            queue_handle,
+            managed: PhantomData,
         }
     }
 
+    /// Start the workers on a managed arbiter, and return the manager struct
+    pub fn start(self) -> Manager {
+        Manager::new(self)
+    }
+}
+
+impl<State> WorkerConfig<State, Unmanaged>
+where
+    State: Clone + 'static,
+{
+    /// Create a new WorkerConfig in the current arbiter
+    ///
+    /// The supplied function should return the State required by the jobs intended to be
+    /// processed. The function must be sharable between threads, but the state itself does not
+    /// have this requirement.
+    pub fn new<S: Storage + Send + Sync + 'static>(
+        storage: S,
+        state_fn: impl Fn(QueueHandle) -> State + Send + Sync + 'static,
+    ) -> Self {
+        Self::new_in_arbiter(Arbiter::current(), storage, state_fn)
+    }
+
+    /// Create a new WorkerConfig in the provided arbiter
+    ///
+    /// The supplied function should return the State required by the jobs intended to be
+    /// processed. The function must be sharable between threads, but the state itself does not
+    /// have this requirement.
+    pub fn new_in_arbiter<S: Storage + Send + Sync + 'static>(
+        arbiter: ArbiterHandle,
+        storage: S,
+        state_fn: impl Fn(QueueHandle) -> State + Send + Sync + 'static,
+    ) -> Self {
+        let queue_handle = create_server_in_arbiter(arbiter.clone(), storage);
+        let q2 = queue_handle.clone();
+
+        WorkerConfig {
+            processors: ProcessorMap::new(Arc::new(move || state_fn(q2.clone()))),
+            queues: BTreeMap::new(),
+            arbiter: Some(arbiter),
+            queue_handle,
+            managed: PhantomData,
+        }
+    }
+
+    /// Start the workers in the provided arbiter
+    pub fn start(self) -> QueueHandle {
+        self.start_managed(self.arbiter.as_ref().unwrap(), &());
+
+        self.queue_handle
+    }
+}
+
+impl<State, M> WorkerConfig<State, M>
+where
+    State: Clone + 'static,
+{
     /// Register a `Job` with the worker
     ///
     /// This enables the worker to handle jobs associated with this processor. If a processor is
@@ -205,31 +405,27 @@ where
         self
     }
 
-    /// Start the workers in the current arbiter
-    ///
-    /// This method will panic if not called from an actix runtime
-    pub fn start(self, queue_handle: QueueHandle) {
-        for (key, count) in self.queues.into_iter() {
-            for _ in 0..count {
-                local_worker(
-                    key.clone(),
-                    self.processors.cached(),
-                    queue_handle.inner.clone(),
-                );
-            }
-        }
-    }
-
-    /// Start the workers in the provided arbiter
-    pub fn start_in_arbiter(self, arbiter: &Arbiter, queue_handle: QueueHandle) {
-        for (key, count) in self.queues.into_iter() {
-            for _ in 0..count {
-                let key = key.clone();
+    /// Start a workers in a managed way
+    fn start_managed<Extras: Clone + Send + 'static>(
+        &self,
+        arbiter: &ArbiterHandle,
+        extras: &Extras,
+    ) {
+        for (key, count) in self.queues.iter() {
+            for _ in 0..*count {
+                let queue = key.clone();
                 let processors = self.processors.clone();
-                let server = queue_handle.inner.clone();
+                let server = self.queue_handle.inner.clone();
+
+                let extras_2 = extras.clone();
 
                 arbiter.spawn_fn(move || {
-                    local_worker(key, processors.cached(), server);
+                    drop(LocalWorkerStarter::new(
+                        queue,
+                        processors.cached(),
+                        server,
+                        extras_2,
+                    ));
                 });
             }
         }
@@ -243,7 +439,6 @@ where
 #[derive(Clone)]
 pub struct QueueHandle {
     inner: Server,
-    arbiter: ArbiterHandle,
 }
 
 impl QueueHandle {
@@ -251,17 +446,12 @@ impl QueueHandle {
     ///
     /// This job will be sent to the server for storage, and will execute whenever a worker for the
     /// job's queue is free to do so.
-    pub fn queue<J>(&self, job: J) -> Result<(), Error>
+    pub async fn queue<J>(&self, job: J) -> Result<(), Error>
     where
         J: Job,
     {
         let job = new_job(job)?;
-        let server = self.inner.clone();
-        self.arbiter.spawn(async move {
-            if let Err(e) = server.new_job(job).await {
-                error!("Error creating job, {}", e);
-            }
-        });
+        self.inner.new_job(job).await?;
         Ok(())
     }
 
@@ -269,17 +459,12 @@ impl QueueHandle {
     ///
     /// This job will be sent to the server for storage, and will execute after the specified time
     /// and when a worker for the job's queue is free to do so.
-    pub fn schedule<J>(&self, job: J, after: DateTime<Utc>) -> Result<(), Error>
+    pub async fn schedule<J>(&self, job: J, after: DateTime<Utc>) -> Result<(), Error>
     where
         J: Job,
     {
         let job = new_scheduled_job(job, after)?;
-        let server = self.inner.clone();
-        self.arbiter.spawn(async move {
-            if let Err(e) = server.new_job(job).await {
-                error!("Error creating job, {}", e);
-            }
-        });
+        self.inner.new_job(job).await?;
         Ok(())
     }
 
@@ -291,7 +476,7 @@ impl QueueHandle {
     where
         J: Job + Clone + Send + 'static,
     {
-        every(self, duration, job);
+        actix_rt::spawn(every(self.clone(), duration, job));
     }
 
     /// Return an overview of the processor's statistics
